@@ -46,6 +46,9 @@ rm_conffile() {
 	[ -n "$1" ] || error "maintainer script parameters are missing"
 	[ -n "$DPKG_MAINTSCRIPT_NAME" ] || \
 		error "environment variable DPKG_MAINTSCRIPT_NAME is required"
+	[ "${CONFFILE}" != "${CONFFILE#/}" ] || \
+		error "conffile '$CONFFILE' is not an absolute path"
+	validate_optional_version "$LASTVERSION"
 
 	debug "Executing $0 rm_conffile in $DPKG_MAINTSCRIPT_NAME" \
 	      "of $DPKG_MAINTSCRIPT_PACKAGE"
@@ -124,7 +127,7 @@ abort_rm_conffile() {
 		mv "$CONFFILE.dpkg-remove" "$CONFFILE"
 	fi
 	if [ -e "$CONFFILE.dpkg-backup" ]; then
-		echo "Reinstalling $CONFFILE that was backupped"
+		echo "Reinstalling $CONFFILE that was backed-up"
 		mv "$CONFFILE.dpkg-backup" "$CONFFILE"
 	fi
 }
@@ -153,6 +156,11 @@ mv_conffile() {
 	[ -n "$1" ] || error "maintainer script parameters are missing"
 	[ -n "$DPKG_MAINTSCRIPT_NAME" ] || \
 		error "environment variable DPKG_MAINTSCRIPT_NAME is required"
+	[ "${OLDCONFFILE}" != "${OLDCONFFILE#/}" ] || \
+		error "old-conffile '$OLDCONFFILE' is not an absolute path"
+	[ "${NEWCONFFILE}" != "${NEWCONFFILE#/}" ] || \
+		error "new-conffile '$NEWCONFFILE' is not an absolute path"
+	validate_optional_version "$LASTVERSION"
 
 	debug "Executing $0 mv_conffile in $DPKG_MAINTSCRIPT_NAME" \
 	      "of $DPKG_MAINTSCRIPT_PACKAGE"
@@ -262,6 +270,7 @@ symlink_to_dir() {
 		error "symlink pathname ends with a slash"
 	[ -n "$SYMLINK_TARGET" ] || error "original symlink target is missing"
 	[ -n "$1" ] || error "maintainer script parameters are missing"
+	validate_optional_version "$LASTVERSION"
 
 	debug "Executing $0 symlink_to_dir in $DPKG_MAINTSCRIPT_NAME" \
 	      "of $DPKG_MAINTSCRIPT_PACKAGE"
@@ -337,6 +346,7 @@ dir_to_symlink() {
 		error "directory parameter is not an absolute path"
 	[ -n "$SYMLINK_TARGET" ] || error "new symlink target is missing"
 	[ -n "$1" ] || error "maintainer script parameters are missing"
+	validate_optional_version "$LASTVERSION"
 
 	debug "Executing $0 dir_to_symlink in $DPKG_MAINTSCRIPT_NAME" \
 	      "of $DPKG_MAINTSCRIPT_PACKAGE"
@@ -404,16 +414,12 @@ prepare_dir_to_symlink()
 
 	# If there are locally created files or files owned by another package
 	# we should not perform the switch.
-	find "$PATHNAME" -print0 | xargs -0 -n1 sh -c '
-		package="$1"
-		file="$2"
-		if ! dpkg-query -L "$package" | grep -F -q -x "$file"; then
-			exit 1
-		fi
-		exit 0
-	' check-files-ownership "$PACKAGE" || \
+	export DPKG_MAINTSCRIPT_HELPER_INTERNAL_API="$version"
+	find "$PATHNAME" -print0 | \
+		xargs -0 -n1 $0 _internal_pkg_must_own_file "$PACKAGE" || \
 		error "directory '$PATHNAME' contains files not owned by" \
 		      "package $PACKAGE, cannot switch to symlink"
+	unset DPKG_MAINTSCRIPT_HELPER_INTERNAL_API
 
 	# At this point, we know that the directory either contains no files,
 	# or only non-conffiles owned by the package.
@@ -483,6 +489,18 @@ abort_dir_to_symlink()
 }
 
 # Common functions
+validate_optional_version() {
+	local VERSION="$1"
+
+	if [ -z "$VERSION" ]; then
+		return
+	fi
+
+	if ! VERSIONCHECK=$(dpkg --validate-version -- "$VERSION" 2>&1); then
+		error "$VERSIONCHECK"
+	fi
+}
+
 ensure_package_owns_file() {
 	local PACKAGE="$1"
 	local FILE="$2"
@@ -495,6 +513,21 @@ ensure_package_owns_file() {
 	return 0
 }
 
+internal_pkg_must_own_file()
+{
+	local PACKAGE="$1"
+	local FILE="$2"
+
+	if [ "$DPKG_MAINTSCRIPT_HELPER_INTERNAL_API" != "$version" ]; then
+		error "internal API used by external command"
+	fi
+
+	if ! ensure_package_owns_file "$PACKAGE" "$FILE"; then
+		error "file '$FILE' not owned by package '$PACKAGE'"
+	fi
+	return 0
+}
+
 symlink_match()
 {
 	local SYMLINK="$1"
@@ -590,6 +623,10 @@ symlink_to_dir)
 dir_to_symlink)
 	dir_to_symlink "$@"
 	;;
+_internal_pkg_must_own_file)
+	# This is an internal command, must not be used outside this program.
+	internal_pkg_must_own_file "$@"
+	;;
 --help|help|-?)
 	usage
 	;;
